import numba
import numpy as np


def ksmooth(x, y, xp, skrn, sbw):
    xp.sort()
    krn = skrn
    bw = sbw
    x = x
    y = y
    nx, nxp = len(x), len(xp)
    yp = np.zeros(nxp, dtype=float)
    o = np.array(sorted(range(0, len(x)), key=lambda i: x[i]))
    return bdr_ksmooth(x.iloc[o,].to_numpy(), y.iloc[o,].to_numpy(), nx, xp, yp, nxp, krn, bw)


@numba.jit(nopython=True)
def bdr_ksmooth(x, y, n, xp, yp, nxp, kern, bw):
    imin = 0
    cutoff = 0.0

    if kern == 1:
        bw *= 0.5
        cutoff = bw
    if kern == 2:
        bw *= 0.3706506
        cutoff = 4 * bw

    while x[imin] < xp[0] - cutoff and imin < n:
        imin += 1

    def dokern(x1, kern1):
        if kern1 == 1:
            return 1.0
        if kern1 == 2:
            return np.exp(-0.5 * x1 * x1)
        return 0.0

    for j in range(nxp):
        num = den = 0.0
        x0 = xp[j]
        for i in range(imin, n):
            if x[i] < x0 - cutoff:
                imin = i
            else:
                if x[i] > x0 + cutoff:
                    break
                w = dokern(np.fabs(x[i] - x0) / bw, kern)
                num += w * y[i]
                den += w
            i += 1
        if den > 0:
            yp[j] = num / den
        else:
            yp[j] = 0
    return xp, yp

# @numba.jit(cache=True, forceobj=True, parallel=True, nogil=True)
# def dokern(x, kern):
#     if kern == 1:
#         return 1.0
#     if kern == 2:
#         return np.exp(-0.5 * x * x)
#     return 0.0

# @numba.jit(cache=True, forceobj=True, parallel=True, nogil=True)
# def BDRksmooth_test(x, y, n, xp, yp, nxp, kern, bw):
#     # start = time.time()
#     imin = 0
#     cutoff = 0.0
#
#     if kern == 1:
#         bw *= 0.5
#         cutoff = bw
#     if kern == 2:
#         bw *= 0.3706506
#         cutoff = 4 * bw
#
#     while x[imin] < xp[0] - cutoff and imin < n:
#         imin += 1
#
#     for j in range(nxp):
#         print(j, nxp, f'{time.time() - start}') if j % 1000 == 0 else ""
#         x0 = xp[j]
#         num, den, imin = test(x0, n, imin, x, cutoff, bw, kern, y)
#         if (den > 0):
#             yp[j] = num / den
#         else:
#             yp[j] = 0
#     # print(f'end cost {time.time() - start}')
#     return xp, yp


# @numba.jit(cache=True, forceobj=True)
# def test(x0, n, imin, x, cutoff, bw, kern, y):
#     num = den = 0.0
#     for i in range(imin, n):
#         if x[i] < x0 - cutoff:
#             imin = i
#         else:
#             if x[i] > x0 + cutoff:
#                 break
#             w = dokern(np.fabs(x[i] - x0) / bw, kern)
#             num += w * y[i]
#             den += w
#         i += 1
#     return num, den, imin


# from joblib import Parallel
# from joblib import delayed
# def BDRksmooth_test(x, y, n, xp, yp, nxp, kern, bw):
#     imin = 0
#     cutoff = 0.0
#
#     if kern == 1:
#         bw *= 0.5
#         cutoff = bw
#     if kern == 2:
#         bw *= 0.3706506
#         cutoff = 4 * bw
#
#     while x[imin] < xp[0] - cutoff and imin < n:
#         imin += 1
#     res = Parallel(n_jobs=8, verbose=100)(
#         delayed(test)(xp[j], n, imin, x, cutoff, bw, kern, y)
#         for j in range(nxp)
#     )
#     yp = np.array(res)
#     return xp, yp
#
# import numba
#
# @numba.jit(cache=True, forceobj=True)
# def test(x0, n, imin, x, cutoff, bw, kern, y):
#     num = den = 0.0
#     for i in range(imin, n):
#         if x[i] < x0 - cutoff:
#             imin = i
#         else:
#             if x[i] > x0 + cutoff:
#                 break
#             w = dokern(np.fabs(x[i] - x0) / bw, kern)
#             num += w * y[i]
#             den += w
#         i += 1
#     return (num / den) if (den > 0) else 0


import random
import time
from typing import Optional
from typing import Union

import numpy as np
from scipy.sparse import csr_matrix

from .scale_data import ScaleData
from .vst import vst

def SCTransform(
        umi: Optional[csr_matrix],
        genes,
        cells,
        reference_sct_model=None,
        do_correct_umi: Optional[bool] = True,
        n_cells: Union[int, None] = 5000,
        residual_features: Union[list, np.ndarray, None] = None,
        variable_features_n: Union[int, None] = 3000,
        variable_features_rv_th: float = 1.3,
        vars_to_regress: Union[str, list, None] = None,
        do_scale: Optional[bool] = False,
        do_center: Optional[bool] = True,
        clip_range=lambda umi: [-np.sqrt(umi.shape[1] / 30), np.sqrt(umi.shape[1] / 30)],
        conserve_memory: Optional[bool] = False,
        return_only_var_genes: Optional[bool] = True,
        seed_use: Union[int, None] = 1448145,
        **kwargs
):
    """
    A Single-Cell RNA Sequencing Transform Method:
        Use regularized negative binomial regression to normalize UMI count data.

    :param umi:
        A csr_matrix of UMI counts with genes as rows and cells as columns
    :param genes:
        Genes describe the `umi`
    :param cells:
        Cells describe the `umi`
    :param reference_sct_model: TODO: only `None` support now.
        If not None, compute residuals for the object using the provided SCT model; supports only log_umi as the latent
        variable. If `residual_features` are not specified, compute for the top variable_features_n specified in the
        model which are also present in the object. If residual_features are specified, the variable features of the
        resultingSCT assay are set to the top variable_features_n in the model.
    :param do_correct_umi: Place corrected UMI matrix in res[0]'s key `counts; default is True
    :param n_cells: Number of sub-sampling cells used to build NB regression; default is 5000
    :param residual_features: TODO not yet finished
        Genes to calculate residual features for; default is None (all genes). If specified, will be set to
        `top_features` of the returned res[1].
    :param variable_features_n:
        Use this many features as variable features after ranking by residual variance; default is 3000. Only applied if
        `residual_features` is not set.
    :param variable_features_rv_th:
        Instead of setting a fixed number of variable features, use this residual variance cutoff; this is only used
        when `variable_features_n` is set to None; default is 1.3. Only applied if `residual_features` is not set.
    :param vars_to_regress: TODO not yet finished
        Variables to regress out in a second non-regularized linear regression. Default is None
    :param do_scale: TODO not yet finished
        Whether to scale residuals to have unit variance; default is False
    :param do_center:
        Whether to center residuals to have mean zero; default is True
    :param clip_range:
        Range to clip the residuals to; default is [-np.sqrt(umi.shape[1] / 30), np.sqrt(umi.shape[1] / 30)], where n
        is the number of cells.
    :param conserve_memory: TODO not yet finished
        If set to True the residual matrix for all genes is never created in full; useful for large data sets, but will
        take longer to run; this will also set `return_only_var_genes` to True; default is False
    :param return_only_var_genes:
        If set to True the scale.data matrices in output res are subset to contain only the variable genes;
        default is True
    :param seed_use:
        Set a random seed. By default, sets the seed to 1448145. Setting None will not set a seed.
    :param kwargs:
        Other arguments, such as `n_genes` defined for `vst`.
    :return:
            {
                'counts':     csr_matrix,       # describe `umi` corrected by pearson residual
                'data':       csr_matrix,       # counts after `log1p`
                'scale.data': pandas.DataFrame, # pearson residual after `scale`
            },
            dict # vst_out
    """
    if seed_use:
        np.random.seed(seed_use)
        random.seed(seed_use)
        print(f'using default random_seed {seed_use}, will run SCT without randomness')

    # different with seurat-SCT, we don't need to choose assay by using exp_matrix directly
    umi = umi.astype(dtype=np.double, copy=False)

    # FIXME: ignore `batch_var`
    if 'batch_var' in kwargs:
        print('`batch_var` not implemented yet, will add the feature in the future')
        raise NotImplementedError

    if callable(clip_range):
        clip_range = clip_range(umi)
    else:
        print(f'clip_range type: {type(clip_range)} is not callable')
        raise TypeError

    vst_args = kwargs
    vst_args['umi'] = umi
    vst_args['genes'] = genes
    vst_args['cells'] = cells
    vst_args['return_cell_attr'] = True
    vst_args['return_gene_attr'] = True
    vst_args['return_corrected_umi'] = do_correct_umi
    # --- modified by @marvinquiet ---
    if n_cells is None:
        vst_args['n_cells'] = umi.shape[1]
    else:
        vst_args['n_cells'] = min(n_cells, umi.shape[1])
    # --- end ---
    vst_args['seed_use'] = seed_use
    # TODO: ignore `res_clip_range` used in 'conserve.memory' sct-method
    # res_clip_range = vst_args['res_clip_range'] if 'res_clip_range' in vst_args else [-math.sqrt(umi.shape[1]),
    #                                                                                  math.sqrt(umi.shape[1])]

    # TODO: ignore `reference_sct_model`、`residual_features` and `conserve.memory`,
    #     we will finish these features in the future (~_~)
    if reference_sct_model:
        # sct_method = 'reference.model'
        raise NotImplementedError
    elif residual_features:
        # sct_method = 'residual.features'
        raise NotImplementedError
    elif conserve_memory:
        # sct_method = 'conserve.memory'
        raise NotImplementedError
    else:
        sct_method = 'default'

    if sct_method == 'default':
        vst_out = vst(**vst_args)
    else:
        raise NotImplementedError

    feature_variance = vst_out['gene_attr']["residual_variance"]
    feature_variance = feature_variance.sort_values(ascending=False)
    vst_out['feature_variance'] = feature_variance
    if variable_features_n:
        top_features = feature_variance[:min(variable_features_n, len(feature_variance))].index.values
    else:
        top_features = feature_variance[feature_variance >= variable_features_rv_th].index.values
    vst_out['top_features'] = top_features

    if sct_method == 'default':
        if return_only_var_genes:
            vst_out['y']['level_0'] = range(0, vst_out['y'].shape[0])
            vst_out['y'] = vst_out['y'].loc[top_features].sort_values('level_0')
            del vst_out['y']['level_0']
    else:
        raise NotImplementedError

    # create output assay
    assay_out = {
        "counts": None,  # sparse matrix
        "data": None,  # sparse matrix
        "scale.data": None  # too dense to generate as sparse matrix
    }

    residual_type = vst_args['residual_type'] if 'residual_type' in vst_args else 'pearson'
    if do_correct_umi and residual_type == 'pearson':
        assay_out["counts"] = vst_out['umi_corrected']
    else:
        assay_out["counts"] = umi

    log1p_counts = assay_out['counts']
    assay_out["data"] = log1p_counts.log1p()

    genes_original_order = vst_out['y'].index.values
    scale_data = vst_out.pop('y')
    scale_data[scale_data < clip_range[0]] = clip_range[0]
    scale_data[scale_data > clip_range[1]] = clip_range[1]

    start_time = time.time()
    scale_data = ScaleData(
        scale_data,
        features=vst_out['top_features'] if return_only_var_genes else None,
        vars_to_regress=vars_to_regress,
        latent_data=vst_out['cell_attr'].T.loc[vars_to_regress] if vars_to_regress else None,
        model_use='linear',
        use_umi=False,
        do_scale=do_scale,
        do_center=do_center,
        scale_max=np.Inf,
        block_size=750,
        min_cells_to_block=3000,
    )
    print(f'scale data cost {time.time() - start_time} seconds')

    if return_only_var_genes:
        assay_out["scale.data"] = scale_data.reindex(genes_original_order, copy=False)
    else:
        assay_out["scale.data"] = scale_data
    return assay_out, vst_out


"""
@marvinquiet:

This package is from Stereopy, originally designed for using `Seurat` flavor SCTransform.
I have made certain modifications for my personal usage.


Coding this python-version SCTransform after learning from these R-packages:
    https://github.com/satijalab/seurat
    https://github.com/satijalab/sctransform

The R-package SCTransform is described in:
    https://genomebiology.biomedcentral.com/articles/10.1186/s13059-019-1874-1
"""
from .sctransform import SCTransform
__all__ = ['SCTransform']



import numba
import numpy as np
import pandas as pd


def ScaleData(
        scale_data: pd.DataFrame,
        features=None,
        vars_to_regress=None,
        latent_data=None,
        split_by=None,
        model_use="linear",
        use_umi=False,
        do_scale=True,
        do_center=True,
        scale_max=10,
        block_size=1000,
        min_cells_to_block=3000,
        verbose=True,
        **kwargs):
    # TODO: `ScaleData` only support one way to run, will fix in the future
    if features is not None:
        features = features
        scale_data = scale_data.loc[features]
    else:
        features = scale_data.index.values

    if vars_to_regress:
        raise NotImplementedError
    max_block = int(np.ceil(len(features) / block_size))
    scaled_data = pd.DataFrame(np.zeros(shape=scale_data.shape, dtype=np.double), index=scale_data.index.values,
                               columns=scale_data.columns.values)
    for i in range(1, max_block + 1):
        my_inds = np.array(range(block_size * (i - 1), block_size * i))
        my_inds = my_inds[my_inds < len(features)]
        arg_list = {
            "mat": scale_data.loc[features[my_inds]],
            "scale": do_scale,
            "center": do_center,
            "scale_max": scale_max,
        }
        scaled_data.loc[features[my_inds]] = fast_row_scale(**arg_list)
    return scaled_data

@numba.jit(cache=True, forceobj=True)
def fast_row_scale(mat, center=True, scale=True, scale_max=10):
    # TODO: ignore `scale`
    rm = None
    if center:
        # the seurat SCT is `float32`
        rm = np.mat(mat.mean(1), dtype=np.float32).T
    if center:
        mat = mat - rm
    if scale_max != np.Inf:
        mat[mat > scale_max] = scale_max
    return mat


@numba.jit(cache=True, forceobj=True)
def fast_row_scale_sparse(mat, center=True, scale=True, scale_max=10):
    # TODO: ignore `scale`
    rm = None
    if center:
        # TODO: `float32` the seurat SCT result
        rm = np.mat(mat.mean(1), dtype=np.float32)
    if center:
        mat = mat - rm
    if scale_max != np.Inf:
        mat[mat > scale_max] = scale_max
    return mat



import numba
import numpy as np
import pandas as pd
from KDEpy import FFTKDE
from joblib import (
    Parallel,
    delayed,
    cpu_count
)
from patsy.highlevel import dmatrix
from scipy import interpolate
from scipy.special import digamma
from scipy.special import polygamma

from .bw import bwSJ

def multi_pearson_residual(i, model_pars_final, regressor_data_final, umi, residual_type, min_variance, genes, bin_ind):
    min_var = min_variance
    genes_bin = genes[bin_ind == i]
    mu = np.exp(np.dot(model_pars_final.loc[genes_bin, ['Intercept', 'log_umi']], regressor_data_final.T))
    genes_bin_bool_list = np.isin(genes, genes_bin)
    y = umi[genes_bin_bool_list, :]
    if residual_type == "pearson":
        return pearson_residual(y, mu, model_pars_final.loc[genes_bin, "theta"], min_var=min_var)
    elif residual_type == "deviance":
        raise NotImplementedError
    else:
        raise Exception


# @numba.jit(cache=True, forceobj=True, nogil=True)
@numba.jit(cache=True, forceobj=True)
def pearson_residual(y, mu, theta, min_var=-np.inf):
    variance = mu + np.divide(mu ** 2, theta.to_numpy().reshape(-1, 1))
    variance[variance < min_var] = min_var
    pearson_residuals = np.divide(y - mu, np.sqrt(variance))
    return pearson_residuals


def dds(genes_log10_gmean_step1, grid_points=2 ** 10):
    x, y = (
        FFTKDE(kernel="gaussian", bw="silverman").fit(np.asarray(genes_log10_gmean_step1)).evaluate(
            grid_points=grid_points)
    )
    density = interpolate.interp1d(x=x, y=y, assume_sorted=False)
    sampling_prob = 1 / (density(genes_log10_gmean_step1) + np.finfo(float).eps)
    return sampling_prob / sampling_prob.sum()


def row_gmean_sparse(umi, gmean_eps=1):
    return np.squeeze(np.asarray(np.array([row_gmean(x.todense(), gmean_eps)[0] for x in umi])))


def row_gmean(umi, gmean_eps=1):
    return np.exp(np.log(umi + gmean_eps).mean(1)) - gmean_eps


def make_cell_attr(umi, cells, latent_var, batch_var, latent_var_nonreg) -> pd.DataFrame:
    # TODO: only `log_umi` will be made, will complete in the future
    tmp_dict_cell_attr = {"umi": umi.sum(0).tolist()[0]}
    tmp_dict_cell_attr["log_umi"] = np.log10(tmp_dict_cell_attr['umi'])
    return pd.DataFrame(tmp_dict_cell_attr, index=cells)


def fit_poisson(umi, model_str, data, theta_estimation_fun="theta.ml") -> pd.DataFrame:
    # TODO: ignore `theta_estimation_fun`
    regressor_data = dmatrix("~log_umi", data, return_type='dataframe')
    results = Parallel(n_jobs=cpu_count(), backend="threading")(
        delayed(one_row_fit_poission)(regressor_data, y.toarray()[0], theta_estimation_fun)
        for y in umi
    )
    return pd.DataFrame(results, columns=["theta", "Intercept", "log_umi"])


@numba.jit(cache=True, forceobj=True, nogil=True)
def one_row_fit_poission(regressor_data, y, theta_estimation_fun='theta.ml'):
    fit = qpois_reg(regressor_data.to_numpy(), y, 1e-9, 100, 1.0001, True)
    if theta_estimation_fun == "theta.ml":
        theta = theta_ml(y=y, mu=fit['fitted'])
    elif theta_estimation_fun == "theta.mm":
        # TODO: `theta.mm` not yet finished
        raise NotImplementedError
    else:
        raise Exception
    return theta, fit['coefficients'][0], fit['coefficients'][1]


@numba.jit(cache=True, forceobj=True, nogil=True)
def qpois_reg(X, Y, tol, maxiters, minphi, returnfit):
    n, pcols = X.shape
    d = pcols

    b_old = np.zeros(shape=(d, 1), dtype=np.double)
    b_new = np.zeros(shape=(d, 1), dtype=np.double)

    y = Y.reshape((Y.shape[0], 1))
    m = np.zeros(shape=(n, 1), dtype=np.double)
    phi = np.zeros(shape=(n, 1), dtype=np.double)

    x = X

    i = 0
    while i < pcols:
        unique_vals = np.unique(x[0:, i])
        if unique_vals.shape[0] == 1:
            b_old[i] = np.log(np.mean(y))
            break
        if unique_vals.shape[0] == 2 and (unique_vals[0] == 0 or unique_vals[1] == 0):
            b_old[i] = np.matmul(y.T, x[0:, i]).item()
            b_old[i] = b_old[i] / np.sum(x[0:, i])
            b_old[i] = np.log(max(1e-9, b_old[i]))
        i += 1

    x_tr = x.T
    ij = 2
    dif = 1.0

    while dif > tol:
        yhat = np.matmul(x, b_old)
        # TODO temporarily value
        yhat = np.clip(yhat, -708, 709)

        m = np.exp(yhat)
        phi = y - m
        L1 = np.matmul(x_tr, phi)
        L2 = []
        for i in range(x.shape[1]):
            L2.append(np.multiply(np.asmatrix(x[0:, i]).T, m))
        L2 = np.array(L2).T
        L2 = np.matmul(x_tr, L2)

        b_new = b_old + np.matmul(np.linalg.inv(L2), L1)[0]
        dif = np.sum(np.abs(b_new - b_old))
        b_old = b_new
        ij += 1
        if ij == maxiters:
            break

    # TODO temporarily value
    phi[(-1.64487933e-154 < phi) & (phi < 1.64487933e-154)] = 0

    p = np.sum(np.square(phi) / m) / (n - pcols)
    return {
        "coefficients": b_new.T[0],
        "phi": p,
        "theta.guesstimate": (np.mean(m) / (max(p, minphi) - 1)),
        "fitted": m.T[0] if returnfit else None
    }


@numba.jit(cache=True, forceobj=True, nogil=True)
def theta_ml(y, mu, limit=10, eps=0.0001220703):
    weights = np.ones(len(y))
    n = np.sum(weights)
    t0 = n / np.sum(np.power(weights * (y / mu - 1), 2))
    it = 1
    _del = 1
    while it < limit and np.abs(_del) > eps:
        t0 = np.abs(t0)
        i = info(n, t0, mu, y, weights)
        _del = score(n, t0, mu, y, weights) / i
        t0 = t0 + _del
        it += 1
    if t0 < 0:
        t0 = 0
    return t0


@numba.jit(cache=True, forceobj=True, nogil=True)
def score(n, th, mu, y, w):
    a = th + y
    b = th + mu
    return np.sum(w * (digamma(a) - digamma(th) + np.log(th) + 1 - np.log(b) - a / b))


# @numba.jit(cache=True, forceobj=True, nogil=True)
@numba.jit(cache=True, forceobj=True)
def trigamma(x):
    return polygamma(1, x)


# @numba.jit(cache=True, forceobj=True, nogil=True)
@numba.jit(cache=True, forceobj=True)
def info(n, th, mu, y, w):
    a = th + y
    b = th + mu
    return np.sum(w * (-trigamma(a) + trigamma(th) - 1 / th + 2 / b - a / np.power(b, 2)))


def is_outlier(y, x, th=10, eps=2.220446e-16 * 10):
    x_max, x_min = x.max(), x.min()
    bin_width = (x_max - x_min) * bwSJ(x) / 2
    breaks1 = seq(x_min - eps, x_max + bin_width, bin_width)
    breaks2 = seq(x_min - eps - bin_width / 2, x_max + bin_width, bin_width)
    score1 = robust_scale_binned(y, x, breaks1)
    score2 = robust_scale_binned(y, x, breaks2)
    return (np.where(abs(score1) < abs(score2), abs(score1), abs(score2)) > th).T[0]


def robust_scale_binned(
        y: pd.DataFrame,
        x,
        breaks
):
    bins = pd.cut(x, bins=breaks, ordered=True)
    y = y.astype(object)
    # TODO: https://pandas.pydata.org/pandas-docs/version/1.5.3/whatsnew/v1.5.1.html
    # changes in `pandas-1.5.1` shows return df whose length more than the length of `bins`
    tmp = y.groupby(bins, observed=True).agg(robust_scale).dropna().explode().to_list()
    o = np.array(sorted(range(0, len(bins)), key=lambda x1: bins[x1]))
    sorted_score = pd.DataFrame(np.zeros(shape=(len(x), 1))).iloc[o,]
    sorted_score[0] = tmp
    return sorted_score.sort_index().values


def mad(x):
    return np.median(np.abs(x - np.median(x))) * 1.4826


def robust_scale(x):
    a = x.values.reshape((x.shape[0],))
    b = np.median(x)
    c = mad(x)
    return (a - b) / (c + 2.220446e-16)


def seq(start, stop, step=1):
    return np.arange(start, stop, step)


import time
from random import sample
from typing import Optional

import numba
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from patsy.highlevel import dmatrix
from scipy.sparse import csr_matrix

from .bw import bwSJ
from .ksmooth import ksmooth
from .sctransform_utils import (
    make_cell_attr,
    cpu_count,
    dds,
    multi_pearson_residual,
    is_outlier,
    fit_poisson,
    row_gmean_sparse
)



def vst(
        umi: Optional[csr_matrix],
        genes,
        cells,
        latent_var='log_umi',
        batch_var=None,
        latent_var_nonreg=None,
        n_genes=2000,
        n_cells=None,
        method='poisson',
        do_regularize=True,
        theta_regularization='od_factor',
        res_clip_range=lambda umi: [-np.sqrt(umi.shape[1]), np.sqrt(umi.shape[1])],
        bin_size=500,
        min_cells=5,
        residual_type='pearson',
        return_cell_attr=False,
        return_gene_attr=True,
        return_corrected_umi=False,
        min_variance=-np.inf,
        bw_adjust=3,
        gmean_eps=1,
        theta_estimation_fun='theta.ml',
        theta_given=None,
        exclude_poisson=False,
        use_geometric_mean=True,
        use_geometric_mean_offset=False,
        fix_intercept=False,
        fix_slope=False,
        scale_factor=None,
        vst_flavor=None,
        seed_use=1448145
):
    # TODO: `vst.flavor` not completed
    if vst_flavor is not None:
        print("`vst.flavor` not completed")
        raise NotImplementedError

    # TODO: `method` not completed, only `poisson`
    if method != 'poisson':
        print("`method` not completed, only `poisson`")
        raise NotImplementedError

    res_clip_range = res_clip_range(umi)
    cell_attr = make_cell_attr(umi, cells, latent_var, batch_var, latent_var_nonreg)
    genes_cell_count = np.asarray((umi >= 0.01).sum(1), dtype=np.double)
    genes_cell_count_pd = pd.DataFrame(genes_cell_count, index=genes, dtype=np.double)
    genes_cell_bool_list = (genes_cell_count.T >= min_cells)[0]
    umi = umi[genes_cell_bool_list, :]
    genes = genes[genes_cell_bool_list,]

    if use_geometric_mean:
        genes_log_gmean = pd.DataFrame(np.log10(row_gmean_sparse(umi, gmean_eps)), index=genes)[0]
    else:
        genes_log_gmean = pd.DataFrame(np.log10(umi.mean(1)), index=genes)[0]

    if do_regularize is False and n_genes:
        n_genes = None
        print("do_regularize is set to False, will use all genes")

    if n_cells and n_cells < umi.shape[1]:
        # down sample cells to speed up the first step
        cells_step1 = sample(cells.tolist(), n_cells)
        cells_step1_bool_list = np.isin(cells, cells_step1)
        genes_cell_count_step1 = umi[:, cells_step1_bool_list].sum(1)
        genes_step1 = genes[(genes_cell_count_step1 >= min_cells).T.tolist()[0],]
        genes_step1_bool_list = np.isin(genes, genes_step1)
        if use_geometric_mean:
            genes_log_gmean_step1 = pd.DataFrame(np.log10(row_gmean_sparse(umi[genes_step1_bool_list, :], gmean_eps)),
                                                 index=genes_step1)[0]
        else:
            genes_log_gmean_step1 = pd.DataFrame(np.log10(umi[genes_step1_bool_list, :].mean(1)), index=genes_step1)[0]
    else:
        cells_step1 = cells
        genes_step1 = genes
        genes_log_gmean_step1 = genes_log_gmean

    genes_amean, genes_var = None, None
    # TODO: `do_regularize` and `exclude_poisson` is True have some code to do something
    #    find over-dispersive genes, this is not default, will be finished
    if do_regularize and exclude_poisson:
        print('finding over-dispersive genes has not finished yet')
        raise NotImplementedError

    if n_genes and n_genes < len(genes_step1):
        sampling_prob = dds(genes_log_gmean_step1)
        genes_step1 = np.array(
            pd.DataFrame(genes_step1).sample(n_genes, weights=sampling_prob.T,
                                             random_state=np.random.RandomState(seed_use))).T[
            0]
        genes_step1_bool_list = np.isin(genes, genes_step1)
        if use_geometric_mean:
            genes_log_gmean_step1 = \
                pd.DataFrame(np.log10(row_gmean_sparse(umi[genes_step1_bool_list, :], gmean_eps)), index=genes_step1)[0]
        else:
            genes_log_gmean_step1 = pd.DataFrame(np.log10(umi[genes_step1_bool_list, :].mean(1)), index=genes_step1)[0]

    print(f'gene-cell umi shape {umi.shape}, n_genes {n_genes} n_cells {n_cells}')

    # TODO: `model_str` need to be recheck, `latent_var` logic not completed
    model_str = "y ~ log_umi"

    data_step1 = cell_attr.loc[cells_step1,]
    cells_step1_bool_list = np.isin(cells, cells_step1)
    genes_step1_bool_list = np.isin(genes, genes_step1)
    start_time = time.time()
    model_pars = get_model_pars(
        genes_step1, bin_size, umi[genes_step1_bool_list,][:, cells_step1_bool_list],
        model_str, cells_step1, method, data_step1, theta_given,
        theta_estimation_fun, exclude_poisson, fix_intercept, fix_slope, use_geometric_mean,
        use_geometric_mean_offset)
    print(f'get_model_pars finished, cost {time.time() - start_time} seconds')

    min_theta = 1e-07
    # model_pars['theta'][model_pars['theta'] < min_theta] = min_theta
    model_pars.loc[model_pars['theta'] < min_theta]['theta'] = min_theta
    model_pars.set_index(genes_step1, inplace=True)

    start_time = time.time()
    if do_regularize:
        model_pars_fit, outliers = reg_model_pars(
            model_pars, genes_log_gmean_step1,
            genes_log_gmean, cell_attr, batch_var, cells_step1,
            genes_step1, umi, bw_adjust, gmean_eps, theta_regularization,
            genes_amean, genes_var, exclude_poisson, fix_intercept,
            fix_slope, use_geometric_mean, use_geometric_mean_offset
        )
    else:
        model_pars_fit, outliers = model_pars, None

    print(f'reg_model_pars finished, cost {time.time() - start_time} seconds')

    regressor_data = dmatrix("~log_umi", cell_attr, return_type='dataframe')
    # TODO: `latent_var_nonreg` not completed yet
    if latent_var_nonreg:
        raise NotImplementedError
    else:
        model_str_nonreg = ""
        model_pars_nonreg = None
        model_pars_final = model_pars_fit
        regressor_data_final = regressor_data

    start_time = time.time()
    if residual_type:
        # TODO `min_variance` not completed related to `vst_flavor`
        bin_ind = np.ceil(np.array(range(1, len(genes) + 1)) / bin_size)
        max_bin = int(np.max(bin_ind))
        res = np.vstack(Parallel(n_jobs=cpu_count(), backend='threading')(
            delayed(multi_pearson_residual)(i, model_pars_final, regressor_data_final, umi, residual_type, min_variance,
                                            genes, bin_ind)
            for i in range(1, max_bin + 1)
        ))
        res = pd.DataFrame(res, index=genes, columns=regressor_data_final.index.values)
    else:
        res = None
    print(f'pearson_residual cost {time.time() - start_time} seconds')

    rv = {
        "y": res,
        "model_str": model_str,
        "model_pars": model_pars,
        "model_outlier": outliers,
        "model_pars_fit": model_pars_fit,
        "model_str_nonreg": model_str_nonreg,
        "model_pars_nonreg": model_pars_nonreg,
        "genes_log_gmean_step1": genes_log_gmean_step1,
        "cells_step1": cells_step1,
        "cell_attr": cell_attr,
        "umi_genes": genes,
        "umi_cells": regressor_data_final.index.values,
    }

    if return_corrected_umi:
        if residual_type == "pearson":
            start_time = time.time()
            rv["umi_corrected"] = correct(rv, genes, do_round=True, do_pos=True, scale_factor=scale_factor,
                                          bin_size=bin_size)
            print(f'umi_corrected cost {time.time() - start_time} seconds')
        else:
            print("will not return corrected UMI because residual type is not set to `pearson`")

    rv["y"][rv["y"] < res_clip_range[0]] = res_clip_range[0]
    rv["y"][rv["y"] > res_clip_range[1]] = res_clip_range[1]
    if not return_cell_attr:
        rv["cell_attr"] = None

    if return_gene_attr:
        gene_attr = pd.DataFrame()
        gene_attr["detection_rate"] = genes_cell_count_pd.loc[genes] / umi.shape[1]
        gene_attr["gmean"] = np.power(10, genes_log_gmean)
        umi_genes_mean = umi.mean(axis=1)
        gene_attr["amean"] = pd.DataFrame(umi_genes_mean, index=genes)
        gene_attr["variance"] = \
            pd.DataFrame(np.power((umi - umi_genes_mean), 2).sum(1) / (umi.shape[1] - 1), index=genes)[0]
        if rv['y'].shape[1] > 0:
            gene_attr["residual_mean"] = rv['y'].mean(1)
            gene_attr["residual_variance"] = rv['y'].var(1)
        rv["gene_attr"] = gene_attr
    return rv


def get_model_pars(genes_step1, bin_size, umi, model_str, cells_step1, method, data_step1, theta_given,
                   theta_estimation_fun, exclude_poisson, fix_intercept, fix_slope, use_geometric_mean,
                   use_geometric_mean_offset) -> pd.DataFrame:
    # TODO: ignore `fix_intercept`、`fix_slope`
    if fix_intercept or fix_slope:
        raise NotImplementedError
    if method.startswith('offset'):
        raise NotImplementedError
    if method == 'poisson':
        model_pars = fit_poisson(umi=umi, model_str=model_str, data=data_step1,
                                 theta_estimation_fun=theta_estimation_fun)
    else:
        raise NotImplementedError
    if exclude_poisson:
        raise NotImplementedError
    return model_pars


def reg_model_pars(
        model_pars,
        genes_log_gmean_step1,
        genes_log_gmean,
        cell_attr,
        batch_var,
        cells_step1,
        genes_step1,
        umi,
        bw_adjust,
        gmean_eps,
        theta_regularization,
        genes_amean=None,
        genes_var=None,
        exclude_poisson=False,
        fix_intercept=False,
        fix_slope=False,
        use_geometric_mean=True,
        use_geometric_mean_offset=False
):
    genes = genes_log_gmean.index.values
    if exclude_poisson or fix_slope or fix_intercept:
        raise NotImplementedError
    dispersion_par = pd.DataFrame()
    if theta_regularization == 'theta':
        dispersion_par[theta_regularization] = np.log10(model_pars["theta"])
    elif theta_regularization == 'od_factor':
        dispersion_par[theta_regularization] = np.log10(1 + np.power(10, genes_log_gmean_step1) / model_pars["theta"])
    else:
        raise Exception(f"{theta_regularization} unknown - only log_theta and od_factor supported at the moment")

    model_pars = model_pars[model_pars.columns[model_pars.columns.values != "theta"]]
    model_pars['dispersion_par'] = dispersion_par.values

    outliers = Parallel(n_jobs=min(cpu_count(), 3), backend="threading")(
        delayed(is_outlier)(col, genes_log_gmean_step1)
        for _, col in model_pars.T.iterrows()
    )

    outliers = np.any(np.array(outliers).T, axis=1)
    if exclude_poisson:
        raise NotImplementedError

    outliers_pd = pd.DataFrame(outliers, dtype=bool, index=model_pars.index.values)
    if np.any(outliers):
        model_pars = model_pars[outliers != True]  # noqa
        # TODO tag genes step1
        # genes_step1 = model_pars.index.values
        genes_log_gmean_step1 = genes_log_gmean_step1[outliers != True]  # noqa

    min_step1, max_step1 = genes_log_gmean_step1.min(), genes_log_gmean_step1.max()
    x_points = np.where(genes_log_gmean > min_step1, genes_log_gmean, min_step1)
    x_points = np.where(x_points < max_step1, x_points, max_step1)
    o = np.array(sorted(range(0, len(x_points)), key=lambda x: x_points[x]))
    model_pars_fit = pd.DataFrame(np.zeros(shape=(len(genes), 3)), index=genes, columns=model_pars.columns.values)
    model_pars_fit = model_pars_fit.reset_index()
    model_pars_fit = model_pars_fit.iloc[o,]

    # TODO: ignore `batch_var`
    start = time.time()
    bw = bwSJ(genes_log_gmean_step1) * bw_adjust
    for col_name, col_value in model_pars.T.iterrows():
        model_pars_fit[col_name] = ksmooth(genes_log_gmean_step1, col_value, x_points.copy(), 2, bw)[1]
    print(f'ksmooth finished, cost {time.time() - start} seconds')

    model_pars_fit = model_pars_fit.sort_index().set_index('index')
    if theta_regularization == 'theta':
        theta = np.power(10, model_pars_fit["dispersion_par"])
    elif theta_regularization == 'od_factor':
        theta = np.power(10, genes_log_gmean) / (np.power(10, model_pars_fit["dispersion_par"]) - 1)
    else:
        raise Exception(f"theta_regularization {theta_regularization} unknown - only log_theta and od_factor supported")
    if exclude_poisson:
        raise NotImplementedError
    if fix_intercept:
        raise NotImplementedError
    if fix_slope:
        raise NotImplementedError
    model_pars_fit = model_pars_fit.loc[:, model_pars_fit.columns != "dispersion_par"]
    model_pars_fit['theta'] = theta
    return model_pars_fit, outliers_pd


def correct(x, genes, as_is=False, do_round=True, do_pos=True, scale_factor=None, bin_size=500):
    if not as_is:
        cell_attr = x['cell_attr'].copy()
        cell_attr['log_umi'] = [np.median(cell_attr['log_umi'])] * len(cell_attr['log_umi'])
    else:
        cell_attr = x['cell_attr']
    regressor_data = dmatrix("~log_umi", cell_attr, return_type='dataframe')
    bin_ind = np.ceil(np.array(range(1, len(genes) + 1)) / bin_size)
    max_bin = int(np.max(bin_ind))
    corrected_data = pd.concat(Parallel(n_jobs=cpu_count(), backend='threading')(
        delayed(multi_correct_data)(x, genes, bin_ind, x['y'], i, regressor_data)
        for i in range(1, max_bin + 1)
    ))
    if do_round:
        corrected_data = np.round(corrected_data, 0)
    if do_pos:
        corrected_data[corrected_data < 0] = 0
    return csr_matrix(corrected_data)


# @numba.jit(cache=True, forceobj=True, nogil=True)
@numba.jit(cache=True, forceobj=True)
def multi_correct_data(x, genes, bin_ind, data, i, regressor_data):
    genes_bin = genes[bin_ind == i]
    pearson_residual_ = data.loc[genes_bin]
    coefs = x['model_pars_fit'].loc[genes_bin, ["Intercept", "log_umi"]]
    theta = x['model_pars_fit'].loc[genes_bin, 'theta']
    return get_correct_data(coefs, regressor_data, theta, pearson_residual_)


# @numba.jit(cache=True, forceobj=True, nogil=True)
@numba.jit(cache=True, forceobj=True)
def get_correct_data(coefs, regressor_data, theta, pearson_residual_):
    mu = np.exp(np.dot(coefs, regressor_data.T))
    variance = mu + np.power(mu, 2) / theta.to_numpy().reshape(-1, 1)
    return mu + pearson_residual_.multiply(np.sqrt(variance))


import os
import scipy
import anndata
import scanpy as sc
import numpy as np
import pandas as pd

from SpaDOT.utils import _utils
from SpaDOT.utils.sctransform import SCTransform

def preprocess_adata(args, adata):
    '''
    Preprocess the data for SpaDOT model training.
    Args:
        args: arguments
        adata: AnnData object
    Returns:
        preprocessed_adata: preprocessed AnnData object
    '''
    tps = adata.obs['timepoint'].unique()
    if args.feature_selection:
        _run_sparkx(args, adata, tps, option='mixture', num_cores=4)
        SVGs = _get_SVGs(args, tps)
        adata = adata[:, SVGs].copy()
    # save counts in adata.layers
    if not scipy.sparse.issparse(adata.X):
        adata.layers['counts'] = scipy.sparse.csr_matrix(adata.X)
    else:
        adata.layers['counts'] = adata.X
    # --- preprocess data
    tp_adata_list = []
    for tp in tps:
        tp_adata = adata[adata.obs['timepoint'] == tp]
        sc.pp.normalize_total(tp_adata, target_sum=1e-4)
        sc.pp.log1p(tp_adata)
        tp_adata_list.append(tp_adata)
    # union SVG genes for all time points
    if args.feature_selection:
        genes = set().union(*(obj.var_names for obj in tp_adata_list))
        genes = list(genes)
        genes.sort()
        with open(args.output_dir+os.sep+'SVG_genes.txt', 'w') as f:
            for item in genes:
                f.write("%s\n" % item)  
    new_tp_adata_list = []
    for tp_adata in tp_adata_list:
        if args.feature_selection:
            tp_adata = tp_adata[:, genes].copy()
        sc.pp.scale(tp_adata)
        new_tp_adata_list.append(tp_adata)
    preprocessed_adata = anndata.concat(new_tp_adata_list)
    return preprocessed_adata


def _get_SVGs(args, tps):
    '''
    Run SPARKX to select SVGs.
    Args:
        args: arguments
    Returns:
        SVG_genes: list of SVG genes
    '''
    # --- after running SPARKX, the SVGs are saved in the result directory
    ## select SVGs by cluster
    tp_SVGs = []
    for tp in tps:
        tp_SVGs_cluster = pd.read_csv(args.output_dir+os.sep+str(tp)+'_SVG_sparkx_clustered_louvain.csv', header=0, index_col=0)
        tp_SVGs.append(tp_SVGs_cluster)
    min_idx = min(range(len(tp_SVGs)), key=lambda i: len(tp_SVGs[i]))
    min_tp_SVG_len = len(tp_SVGs[min_idx])
    SVG_genes = tp_SVGs[min_idx].index.tolist()
    for idx, tp_SVG in enumerate(tp_SVGs):
        if idx == min_idx: continue
        tp_SVG_num_clusters = len(set(tp_SVG['cluster']))
        tp_SVG_top = tp_SVG.sort_values(by='adjustedPval', ascending=True).groupby('cluster').head(max(100, round(min_tp_SVG_len/tp_SVG_num_clusters)))
        SVG_genes.extend(tp_SVG_top.index.tolist())
    SVG_genes = list(set(SVG_genes))
    SVG_genes.sort()
    return SVG_genes


def _run_sparkx(args, adata, tps, option='mixture', num_cores=4):
    '''
    Run SPARKX-X python version and write results into files.
    '''
    for tp in tps:
        tp_adata = adata[adata.obs['timepoint'] == tp].copy()
        tp_adata.layers['counts'] = tp_adata.X.copy()
        assay_out, vst_out = SCTransform(tp_adata.X.T,
                            genes=tp_adata.var_names,
                            cells=tp_adata.obs_names,
                            return_only_var_genes=False, 
                            n_cells=None,
                            variable_features_n=None,
                            variable_features_rv_th=1.3)
        tp_adata = tp_adata[:, assay_out['scale.data'].index].copy()
        print(f'Timepoint: {tp}, Number of cells: {tp_adata.n_obs}, Number of genes: {tp_adata.n_vars}')
        count_spark = tp_adata.layers['counts']
        locations_spark = tp_adata.obsm['spatial']
        SVGs = _utils._sparkx(count_spark, locations_spark, np.array(tp_adata.var_names), option=option, num_cores=num_cores)
        # SVGs.to_csv(args.output_dir+os.sep+str(tp)+'_SVG_sparkx.csv')
        SVG_clusters = _utils._cluster_SVGs(assay_out['scale.data'].loc[SVGs.index, :], k=10)
        SVGs['cluster'] = SVG_clusters
        SVGs.to_csv(args.output_dir+os.sep+str(tp)+'_SVG_sparkx_clustered_louvain.csv')


# --- other utilitities functions ---
import os
import time
import yaml
# from importlib.resources import files
import sklearn
import scipy as sp
import numpy as np
import pandas as pd
import scanpy as sc
import torch
import random
from torch.backends import cudnn

import warnings
from scipy.stats import cauchy
from multiprocessing import Pool
from scipy.stats import ncx2
from chi2comb import chi2comb_cdf, ChiSquared # implementation of davies function in r: https://www.rdocumentation.org/packages/CompQuadForm/versions/1.4.4/topics/davies
from statsmodels.stats.multitest import multipletests

def set_seed(seed=1993):
    os.environ['PYTHONHASHSEED'] = str(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    cudnn.deterministic = True
    cudnn.benchmark = False
    g = torch.Generator()
    g.manual_seed(seed) # for stabilization

def seed_worker(worker_id=1993):
    np.random.seed(worker_id)
    random.seed(worker_id)

def load_model_config(args):
    '''
    Load the model configuration from the config.yaml file provided in the package.
    '''
    if args.config:
        config_path = args.config
    else:
        # Load the default config.yaml from the package
        # config_path = files("SpaDOT").joinpath("config.yaml")
        config_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "config.yaml")
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)
    return config

def _Cal_Spatial_Net(adata, k_cutoff=None, max_neigh=30):
    """
    Construct the spatial neighbor networks using KNN.
    Parameters

    adata : anndata.AnnData
        The AnnData object containing the spatial data.
    """

    print('Calculating spatial graph...')
    coor = pd.DataFrame(adata.obsm['spatial'])
    coor.index = adata.obs.index
    coor.columns = ['imagerow', 'imagecol']

    nbrs = sklearn.neighbors.NearestNeighbors(
        n_neighbors=max_neigh + 1, algorithm='auto').fit(coor)
    distances, indices = nbrs.kneighbors(coor)
    indices = indices[:, 1:k_cutoff + 1]
    distances = distances[:, 1:k_cutoff + 1]

    KNN_list = []
    for it in range(indices.shape[0]):
        KNN_list.append(pd.DataFrame(zip([it] * indices.shape[1], indices[it, :], distances[it, :])))
    KNN_df = pd.concat(KNN_list)
    KNN_df.columns = ['Cell1', 'Cell2', 'Distance']

    Spatial_Net = KNN_df.copy()
    id_cell_trans = dict(zip(range(coor.shape[0]), np.array(coor.index), ))
    Spatial_Net['Cell1'] = Spatial_Net['Cell1'].map(id_cell_trans)
    Spatial_Net['Cell2'] = Spatial_Net['Cell2'].map(id_cell_trans)

    print('The graph contains %d edges, %d cells.' % (Spatial_Net.shape[0], adata.n_obs))
    print('%.4f neighbors per cell on average.' % (Spatial_Net.shape[0] / adata.n_obs))
    adata.uns['Spatial_Net'] = Spatial_Net

    X = pd.DataFrame(adata.layers['counts'].toarray()[:, ], index=adata.obs.index, columns=adata.var.index)
    cells = np.array(X.index)
    cells_id_tran = dict(zip(cells, range(cells.shape[0])))
    if 'Spatial_Net' not in adata.uns.keys():
        raise ValueError("Spatial_Net is not existed! Run Cal_Spatial_Net first!")
        
    # add self-loop
    Spatial_Net = adata.uns['Spatial_Net']
    G_df = Spatial_Net.copy()
    G_df['Cell1'] = G_df['Cell1'].map(cells_id_tran)
    G_df['Cell2'] = G_df['Cell2'].map(cells_id_tran)
    G = sp.sparse.coo_matrix((np.ones(G_df.shape[0]), (G_df['Cell1'], G_df['Cell2'])), shape=(adata.n_obs, adata.n_obs))
    G = G + np.eye(G.shape[0])  # add self-loop
    adata.uns['adj'] = G

def _save_inducing_points(args, inducing_points_dict):
    """
    Write the inducing points to a file.
    """
    # Convert the dictionary into a DataFrame directly
    inducing_points_list = []
    for key, value in inducing_points_dict.items():
        # Create a DataFrame for each key-value pair
        df = pd.DataFrame(value)
        df.columns = ['norm-pixel_x', 'norm-pixel_y']
        df["timepoint"] = key  # Add the key as a new column
        inducing_points_list.append(df)
    # Combine all DataFrames into a single DataFrame
    inducing_points_df = pd.concat(inducing_points_list, ignore_index=True)

    # Save to a file
    inducing_points_df.to_csv(args.output_dir+os.sep+args.prefix+"inducing_points.csv", index=False)


def _sparkx(count, location, genenames, option='mixture', num_cores=1):
    """Run SPAKR-X python version
    Parameters
    ----------
    count : scipy.sparse matrix or np.ndarray
        Gene expression count matrix (cells x genes).
    location : np.ndarray
        Spatial coordinates matrix (cells x dimensions).
    genenames : np.ndarray
        Array of gene names corresponding to count columns. 
    option : str, optional
        Type of kernel to use ('mixture' or 'projection'). Default is 'mixture'.
    num_cores : int, optional
        Number of CPU cores to use for parallel processing. Default is 1.
    """
    assert count.shape[1] == len(genenames), "Length of genenames must match number of columns in count"
    
    # --- perform filtering on cells / genes
    totalcount = count.sum(axis=1)
    keep_cell_idx = np.where(totalcount != 0)[0]
    count = count[keep_cell_idx, :]
    location = location[keep_cell_idx, :]
    genecount = np.array(count.sum(axis=0)).ravel()
    keep_gene_idx = np.where(genecount != 0)[0]
    count = count[:, keep_gene_idx]
    genenames = genenames[keep_gene_idx]
    if np.sum(pd.isna(genenames)) > 0:
        genenames[pd.isna(genenames)] = "NAgene"
    numGene = count.shape[1]
    numCell = count.shape[0]
    # --- run SPARKX ---
    print("## ===== SPARK-X INPUT INFORMATION ==== ")
    print(f"## number of total samples: {numCell}")
    print(f"## number of total genes: {numGene}")
    if num_cores > 1:
        print(f"## Running with {num_cores} cores")
    else:
        print("## Running with single core, may take some time")
    sparkx_list = list()
    print("## Testing With Projection Kernel")
    final_location = location
    sparkx_res = _sparkx_sk(count, final_location, num_cores=num_cores)
    sparkx_res.index = genenames
    sparkx_list.append(sparkx_res)
    if option == 'mixture':
        # --- record run time
        start_time = time.time()
        for iker in range(5):
            print(f"## Testing With Gaussian Kernel {iker+1}")
            final_location = _transloc_func_vec(location, lker=iker, transfunc="gaussian")
            sparkx_list.append(_sparkx_sk(count, final_location, num_cores=num_cores))
        for iker in range(5):
            print(f"## Testing With Cosine Kernel {iker+1}")
            final_location = _transloc_func_vec(location, lker=iker, transfunc="cosine")
            sparkx_list.append(_sparkx_sk(count, final_location, num_cores=num_cores))
        end_time = time.time()
        print(f"Time taken for mixture kernels: {end_time - start_time:.2f} seconds")
    # --- combine p-values from different kernels using ACAT ---
    allstat = np.column_stack([x["stat"] for x in sparkx_list])
    allpvals = np.column_stack([x["pval"] for x in sparkx_list])
    allstat = pd.DataFrame(allstat, index=genenames)    # shape: n_genes × n_methods
    allpvals = pd.DataFrame(allpvals, index=genenames)
    comb_pval = allpvals.apply(_ACAT, axis=1)
    pBY = multipletests(comb_pval, method="fdr_by")[1]
    res_sparkx = pd.DataFrame({
        'combinedPval': comb_pval,
        'adjustedPval': pBY
    }, index=genenames)
    res_sparkx = res_sparkx.sort_values(by='adjustedPval')
    significant_gene_number = (res_sparkx['adjustedPval'] <= 0.05).sum()
    significant_gene_number = min(res_sparkx.shape[0], max(significant_gene_number, 500))  # avoid too few genes
    SVGs = res_sparkx.sort_values('adjustedPval').iloc[:significant_gene_number, :]
    return SVGs

def _cluster_SVGs(SVG_mat, k=10):
    # --- obtain result from Scanpy
    # record cluster time
    start_time = time.time()
    SVG_adata = sc.AnnData(X=SVG_mat)
    sc.tl.pca(SVG_adata)
    sc.pp.neighbors(SVG_adata, n_neighbors=100, n_pcs=30, method='gauss')
    # sc.pp.neighbors(SVG_adata, n_neighbors=100, n_pcs=30)
    # D = SVG_adata.obsp["distances"].tocoo()
    # W_data = 1.0 / (1.0 + D.data)
    # W = sp.sparse.csr_matrix((W_data, (D.row, D.col)), shape=D.shape)
    # W = W + W.T # make undirected
    # SVG_adata.obsp["connectivities"] = W
    resolution = 1.0
    sc.tl.louvain(SVG_adata,
                 resolution=resolution, 
                 key_added=f"louvain_{resolution}",
                 use_weights=True)
    while len(set(SVG_adata.obs[f"louvain_{resolution}"])) < k:
        resolution += 0.1
        sc.tl.louvain(SVG_adata, 
                    resolution=resolution,
                    key_added=f"louvain_{resolution}",
                    use_weights=True)
    end_time = time.time()
    print(f"Time taken for clustering SVGs: {end_time - start_time:.2f} seconds")
    return SVG_adata.obs[f"louvain_{resolution}"].values

def _sparkx_sk(counts, infomat, num_cores=1):
    """Simplified SPARK-X without covariate matrix
    """
    Xinfomat = infomat - infomat.mean(axis=0, keepdims=True)
    XtX = Xinfomat.T @ Xinfomat
    loc_inv = np.linalg.inv(XtX)
    kmat_first = Xinfomat @ loc_inv
    Klam = np.linalg.eigvalsh(Xinfomat.T @ kmat_first) # guarantees real eigenvalues if the matrix is close to symmetric.
    EHL = counts.T @ Xinfomat
    numCell = Xinfomat.shape[0]
    counts_squared = counts.power(2)
    adjust_nominator = np.array(counts_squared.sum(axis=0)).ravel()
    vec_stat = np.einsum('ij,jk,ik->i', EHL, loc_inv, EHL)  # shape: (n_rows,)
    # scale by numCell and divide by adjust_nominator
    vec_stat = vec_stat * numCell / adjust_nominator
    vec_ybar = np.array(counts.mean(axis=0)).ravel()
    vec_ylam = 1 - numCell * vec_ybar**2 / adjust_nominator  # vectorized
    # Compute davies p-value in parallel ---
    with Pool(num_cores) as pool:
        vec_daviesp = pool.starmap(
            _sparkx_pval,
            [(i, vec_ylam, Klam, vec_stat) for i in range(counts.shape[1])]
        )    
    vec_daviesp = np.array(vec_daviesp)  # convert to NumPy array
    res_sparkx = pd.DataFrame({
        'stat': vec_stat,
        'pval': vec_daviesp
    })
    return res_sparkx

def _sparkx_pval(igene,lambda_G,lambda_K,allstat):
    '''
    Calculate p-values for SPARKX

    Parameters
    ----------
    igene: int
        Gene index
    lambda_G : array-like
        Y (lambda)
    lambda_K : array-like
        Eigenvalue of the gene (lambda)
    allstat : array-like
        Statistics for calculating p-values

    Returns
    -------
    float
        P-value P[Q > q]
    '''
    try:
        # sort lambda_G[igene] * lambda_K in decreasing order
        Zsort = np.sort(lambda_G[igene] * lambda_K)[::-1]
        gcoef = 0
        dofs = [1.0] * len(Zsort)
        ncents = [0.0] * len(Zsort)
        q = allstat[igene]
        chi2s = [ChiSquared(Zsort[i], ncents[i], dofs[i]) for i in range(len(Zsort))]
        result, errno, info = chi2comb_cdf(q, chi2s, gcoef)
        pout = 1 - result
        if result <= 0 or result >= 1.0:
            pout = _liu(allstat[igene], Zsort)
    except Exception:
        pout = _liu(allstat[igene], Zsort)
    return pout


def _liu(q, lambdas, h=None, delta=None):
    """ -> converted by ChatGPT, original code from https://github.com/cran/CompQuadForm/blob/master/R/liu.R
    Liu approximation for the distribution of quadratic forms
    in normal variables (R 'liu' function).

    Parameters
    ----------
    q : float
        The value to evaluate P[Q > q]
    lambdas : array-like
        Eigenvalues (lambda)
    h : array-like, optional
        Multiplicities (default 1 for each lambda)
    delta : array-like, optional
        Non-centrality parameters (default 0 for each lambda)

    Returns
    -------
    float
        P-value P[Q > q]
    """
    
    lambdas = np.asarray(lambdas)
    r = len(lambdas)
    if h is None:
        h = np.ones(r)
    else:
        h = np.asarray(h)
        if len(h) != r:
            raise ValueError("lambda and h should have the same length!")
    if delta is None:
        delta = np.zeros(r)
    else:
        delta = np.asarray(delta)
        if len(delta) != r:
            raise ValueError("lambda and delta should have the same length!")
        if np.any(delta < 0):
            raise ValueError("All non centrality parameters in 'delta' should be positive!")
    c1 = np.sum(lambdas * h) + np.sum(lambdas * delta)
    c2 = np.sum(lambdas**2 * h) + 2 * np.sum(lambdas**2 * delta)
    c3 = np.sum(lambdas**3 * h) + 3 * np.sum(lambdas**3 * delta)
    c4 = np.sum(lambdas**4 * h) + 4 * np.sum(lambdas**4 * delta)
    s1 = c3 / (c2 ** (3 / 2))
    s2 = c4 / (c2 ** 2)
    muQ = c1
    sigmaQ = np.sqrt(2 * c2)
    tstar = (q - muQ) / sigmaQ
    if s1**2 > s2:
        a = 1 / (s1 - np.sqrt(s1**2 - s2))
        delta = s1 * a**3 - a**2
        l = a**2 - 2 * delta
    else:
        a = 1 / s1
        delta = 0
        l = c2**3 / c3**2
    muX = l + delta
    sigmaX = np.sqrt(2) * a
    # Survival function (upper tail)
    try:
        Qq = ncx2.sf(tstar * sigmaX + muX, df=l, nc=delta) 
    except Exception:
        print("Error in computing survival function")
    return Qq

def _ACAT(pvals, weights=None):
    # --- converted by ChatGPT, original code from: https://github.com/xzhoulab/SPARK/blob/a8b4bf27b804604dfda53da42992f100b8e4e727/R/sparkx.R#L307
    
    # check for NAs
    if np.any(np.isnan(pvals)):
        raise ValueError("Cannot have NAs in the p-values!")
    # check if pvals are between 0 and 1
    if np.any(pvals < 0) or np.any(pvals > 1):
        raise ValueError("P-values must be between 0 and 1!")
    # check if there are pvals that are exactly 0 or 1
    is_zero = np.any(pvals == 0)
    is_one = np.any(pvals == 1)
    if is_zero and is_one:
        raise ValueError("Cannot have both 0 and 1 p-values!")
    if is_zero:
        return 0.0
    if is_one:
        warnings.warn("There are p-values that are exactly 1!")
        return 1.0
    # default equal weights, or normalize user-supplied weights
    n = len(pvals)
    if weights is None:
        weights = np.ones(n) / n
    else:
        weights = np.array(weights)
        if len(weights) != n:
            raise ValueError("The length of weights should be the same as that of the p-values")
        if np.any(weights < 0):
            raise ValueError("All the weights must be positive!")
        weights = weights / np.sum(weights)
    # handle very small p-values
    is_small = pvals < 1e-16
    if not np.any(is_small):
        cct_stat = np.sum(weights * np.tan((0.5 - pvals) * np.pi))
    else:
        cct_stat = np.sum(weights[is_small] / (np.pi * pvals[is_small]))
        cct_stat += np.sum(weights[~is_small] * np.tan((0.5 - pvals[~is_small]) * np.pi))
    # compute p-value
    if cct_stat > 1e15:
        pval = 1 / (cct_stat * np.pi)
    else:
        pval = 1 - cauchy.cdf(cct_stat)
    return pval

def _transloc_func_vec(coord, lker, transfunc="gaussian"):
    """
    Transform locations
    """
    # center each column
    coord = coord - np.mean(coord, axis=0)
    # compute quantiles per column
    probs = np.arange(0.2, 1.01, 0.2)
    # l will be a 2D array: each column has its quantiles
    l = np.quantile(np.abs(coord), q=probs, axis=0)
    if transfunc == "gaussian":
        out = np.exp(-coord**2 / (2 * l[lker, :][np.newaxis, :]**2))
    elif transfunc == "cosine":
        out = np.cos(2 * np.pi * coord / l[lker, :][np.newaxis, :])
    else:
        raise ValueError("transfunc must be 'gaussian' or 'cosine'")
    return out
